from dataclasses import dataclass


@dataclass
class EnvironmentDefinition:
    uid: str
    environment_name: str
    environment_params: dict
    environment_uid: str
    reward_name: str
    reward_params: dict
    state_transformer_name: str
    state_transformer_params: dict



import logging
import os
from typing import Any, Dict, List, Union

from arsenai.generator import Experiment, Generator
from arsenai.util import create_run_files

LOG = logging.getLogger("arsenai.api")


def generate(
    experiment_file: Union[str, os.PathLike], no_disk_saving: bool = False
) -> List[Dict[str, Any]]:
    LOG.info(f"Reading experiment file: {experiment_file}.")

    experiment = Experiment.load(experiment_file)
    generator = Generator()
    runs, design = generator.generate(experiment)

    if not no_disk_saving:
        create_run_files(runs, experiment.uid, experiment.output_path)
        design.to_csv(
            os.path.join(
                experiment.output_path, f"{experiment.uid}_design.csv"
            )
        )

    LOG.info("ArsenAI finished!")
    return runs



import unittest
from copy import copy

import numpy as np
from arsenai.description import ParameterDefinition
from arsenai.generator import designer


class TestDesigner(unittest.TestCase):
    def setUp(self):
        ind_vars = {
            "0.environments._": {0: "test_env1", 1: "test_env2"},
            "0.agents._": {0: "test_agent1", 1: "test_agent2"},
            "0.sensors.test_agent1": {
                0: "minimal",
                1: "medium",
                2: "even_more",
                3: "all",
            },
        }
        self.pardef = ParameterDefinition()
        for key, value in ind_vars.items():
            self.pardef.add_independent_variable(key, value)

    def test_create_design_enough_runs(self):
        design = designer.create_design(self.pardef, 16)

        self.assertEqual(16, len(design))

    def test_create_design_not_enough_runs(self):
        design = designer.create_design(self.pardef, 4)

        self.assertEqual(4, len(design))

    def test_optimize(self):
        original = np.array(
            [
                [0.0, 3.0, 0.0],
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [3.0, 2.0, 3.0],
            ]
        )

        design = np.array(
            [
                [1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0],
            ]
        )

        self.assertTrue(
            designer._isbetter(
                designer._optimize(design, original, 1000), design
            )
        )

    def test_isbetter(self):
        design1 = np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [2.0, 0.0, 0.0],
            ]
        )

        design2 = np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 1.0, 1.0],
                [3.0, 3.0, 3.0],
            ]
        )

        self.assertTrue(designer._isbetter(design2, design1))

    def test_mutate(self):
        original = np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [2.0, 0.0, 1.0],
                [0.0, 1.0, 0.0],
                [1.0, 2.0, 3.0],
            ]
        )

        design = np.array(
            [
                [1.0, 0.0, 0.0],
                [2.0, 0.0, 1.0],
                [0.0, 1.0, 0.0],
            ]
        )

        design_copy = copy(design)
        designer._mutate(design_copy, original)
        self.assertEqual(design.shape, design_copy.shape)
        self.assertTrue(np.max(design_copy[:, 0]) <= 2)
        self.assertTrue(np.max(design_copy[:, 1]) <= 2)
        self.assertTrue(np.max(design_copy[:, 1]) <= 3)

    def test_pdist(self):
        design = np.array(
            [
                [0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0],
                [2.0, 0.0, 1.0],
                [0.0, 1.0, 0.0],
                [1.0, 2.0, 3.0],
            ]
        )
        distences = designer._pdist(design)
        distences_expected = np.array(
            [
                1.0,
                2.23606797749979,
                1.0,
                3.7416573867739413,
                1.4142135623730951,
                1.4142135623730951,
                3.605551275463989,
                2.449489742783178,
                3.0,
                3.3166247903554,
            ]
        )
        self.assertEqual(distences.tolist(), distences_expected.tolist())


if __name__ == "__main__":
    unittest.main()



# from arsenai.schedule.phase import Phase

from . import LOG
from .phase import Phase


class Schedule:
    """This class stores information about the schedule for a specific
    experiment run.

    """

    def __init__(self, blueprint: list):
        self._blueprint: list = blueprint
        self.phases: list = list()

    def populate(self):
        for phase_idx, phase_def in enumerate(self._blueprint):
            name = list(phase_def.keys())[0]
            phase = Phase(phase_idx, name, phase_def[name])
            phase.populate()
            self.phases.append(phase)

    def build(self, experiment, design_idx):
        for phase in self.phases:
            for request in phase.requests:
                LOG.debug("Processing request: %s", request)
                prev_phase = self.phases[request["from_phase"]]
                attr = getattr(phase, request["what"])
                if "entry" in request:
                    attr[request["level"]][request["entry"]] = getattr(
                        prev_phase, request["what"]
                    )[request["index"]]
                else:
                    prev_attr = getattr(prev_phase, request["what"])
                    if isinstance(prev_attr, list):
                        attr[request["level"]] = prev_attr[request["level"]]
                    else:
                        setattr(phase, request["what"], prev_attr)
            phase.build(experiment, design_idx)

    def as_list(self):
        schedule = list()
        for phase in self.phases:
            schedule.append(phase.as_dict())
        return schedule



import logging

LOG = logging.getLogger(__name__)

from .actuator_definition import ActuatorDefinition
from .agent_definition import AgentDefinition
from .parameter_definition import ParameterDefinition
from .environment_definition import EnvironmentDefinition
from .phase_config_definition import PhaseConfigDefinition
from .run_config_definition import RunConfigDefinition
from .sensor_definition import SensorDefinition
from .simulation_definition import SimulationDefinition



import os
import shutil
import unittest

import ruamel.yaml as yml
from arsenai.api import fnc_generate


class TestCase01SingleRun(unittest.TestCase):
    def setUp(self) -> None:
        fixtures_path = os.path.abspath(
            os.path.join(__file__, "..", "..", "fixtures")
        )
        self.experiment_file = os.path.join(
            fixtures_path, "testcase01_single_run.yml"
        )
        self.expected_run_file = os.path.join(
            fixtures_path, "testcase01_single_run_expected.yml"
        )
        self.tmp_path = os.path.abspath(os.path.join(os.getcwd(), "tmp"))
        self.actual_run_file = os.path.join(
            self.tmp_path, "TestCase01-SingleRun_run-0.yml"
        )

    def test_generate(self):
        with open(self.expected_run_file, "r") as yml_file:
            expected = yml.YAML(typ="safe", pure=True).load(yml_file)

        try:
            fnc_generate.generate(self.experiment_file)
        except SystemExit:
            pass

        with open(self.actual_run_file, "r") as yml_file:
            actual = yml.YAML(typ="safe", pure=True).load(yml_file)

        self.maxDiff = None
        self.assertDictEqual(expected, actual)

    def tearDown(self):
        shutil.rmtree(self.tmp_path)


if __name__ == "__main__":
    unittest.main()



from arsenai.cli.arsenai import main


if __name__ == "__main__":
    main(
        [
            "-vv",
            "generate",
            "/home/stephan/data/work/palaestrai/"
            "arsenai/tests/fixtures/example_experiment.yml",
        ]
    )



"""This module contains the experiment description for arsenAI."""

import os
from copy import deepcopy
from typing import List

from palaestrai.util import seeding

from ..description import (
    ActuatorDefinition,
    AgentDefinition,
    EnvironmentDefinition,
    PhaseConfigDefinition,
    RunConfigDefinition,
    SensorDefinition,
    SimulationDefinition,
)
from ..util import load_experiment_file
from . import LOG


class Experiment:
    """This object holds information about a single experiment."""

    def __init__(
        self,
        uid: str,
        version: str,
        seed: int,
        repetitions: str,
        max_runs: int,
        output_path: str,
        environment_definitions: List[EnvironmentDefinition],
        agent_definitions: List[AgentDefinition],
        sensor_definitions: List[SensorDefinition],
        actuator_definitions: List[ActuatorDefinition],
        simulation_definitions: List[SimulationDefinition],
        phase_config_definitions: List[PhaseConfigDefinition],
        run_config_definition: RunConfigDefinition,
        schedule: list,
    ):
        self.uid = uid
        (
            self.version_major,
            self.version_minor,
            self.version_patch,
        ) = (int(val) for val in version.split("."))
        self.rng, self.seed = seeding.np_random(seed)
        self.repetitions = repetitions
        self.max_runs = max_runs
        self.output_path = output_path
        self.environment_definitions = environment_definitions
        self.agent_definitions = agent_definitions
        self.sensor_definitions = sensor_definitions
        self.actuator_definitions = actuator_definitions
        self.simulation_definitions = simulation_definitions
        self.phase_config_definitions = phase_config_definitions
        self.run_config_definition = run_config_definition
        self.schedule = schedule

    def get_seed(self):
        # LOG.warning(
        #     "Reusing experiment seed for run! "
        #     "See arsenai.generator.experiment.py line 61.",
        # )
        return self.rng.randint(1_000_000)

    def get_version(self):
        return (
            f"{self.version_major}.{self.version_minor}.{self.version_patch}"
        )

    def get_schedule(self):
        return deepcopy(self.schedule)

    def get_run_config(self):
        return {
            "condition": {
                "name": self.run_config_definition.condition_name,
                "params": deepcopy(
                    self.run_config_definition.condition_params
                ),
            }
        }

    def get_definition(self, topic, user_key, phase_name):
        if topic == "environments":
            return self.get_environment_definition(user_key)
        if topic == "agents":
            return self.get_agent_definition(user_key, phase_name)
        if topic == "simulation":
            return self.get_simulation_definition(user_key)
        if topic == "phase_config":
            return self.get_phase_config_definition(user_key)
        if topic == "sensors":
            return self.get_sensor_definition(user_key)
        if topic == "actuators":
            return self.get_actuator_definition(user_key)

        raise KeyError(f"Unknown topic: '{topic}'")

    def get_environment_definition(self, uids):
        defs = list()
        for uid in uids:
            for envdef in self.environment_definitions:
                if envdef.uid == uid:
                    env_cfg = {
                        "environment": {
                            "name": envdef.environment_name,
                            "params": deepcopy(envdef.environment_params),
                            "uid": envdef.environment_uid,
                        },
                    }

                    # Reward may be emtpy
                    if envdef.reward_name:
                        env_cfg["reward"] = {
                            "name": envdef.reward_name,
                            "params": envdef.reward_params,
                        }

                    # State transformer may be empty
                    if envdef.state_transformer_name:
                        env_cfg["state_transformer"] = {
                            "name": envdef.state_transformer_name,
                            "params": envdef.state_transformer_params,
                        }
                    defs.append(env_cfg)
                    break
        return defs

    def get_agent_definition(self, uids, phase_name):
        defs = list()
        for uid in uids:
            for adef in self.agent_definitions:
                if adef.uid == uid:
                    data = {
                        "uid": adef.uid,
                        "name": adef.name,
                        "brain": {
                            "name": adef.brain_name,
                            "params": deepcopy(adef.brain_params),
                        },
                        "muscle": {
                            "name": adef.muscle_name,
                            "params": adef.muscle_params,
                        },
                        "objective": {
                            "name": adef.objective_name,
                            "params": adef.objective_params,
                        },
                        "load": (
                            adef.load[phase_name]
                            if phase_name in adef.load
                            else adef.load
                        ),
                        "replay": adef.replay,
                    }
                    try:
                        load = adef.load[phase_name]
                        if load:
                            data["load"] = load
                    except TypeError:
                        pass  # No load definition
                    except KeyError:
                        pass  # No load definition

                    defs.append(data)
                    break
        if len(defs) != len(uids):
            LOG.warning(
                "Could not determine all required configs: %s --> %s",
                uids,
                defs,
            )
        return defs

    def get_simulation_definition(self, uid):
        if isinstance(uid, list):
            uid = uid[0]
        for sdef in self.simulation_definitions:
            if sdef.uid == uid:
                return {
                    "name": sdef.simulation_name,
                    "conditions": deepcopy(sdef.conditions),
                }

    def get_phase_config_definition(self, uid):
        if isinstance(uid, list):
            uid = uid[0]
        for pcdef in self.phase_config_definitions:
            if pcdef.uid == uid:
                return {
                    "mode": pcdef.mode,
                    "worker": pcdef.worker,
                    "episodes": pcdef.episodes,
                }

    def get_sensor_definition(self, uid):
        for sendef in self.sensor_definitions:
            if sendef.uid == uid:
                return sendef.sensor_ids

        raise KeyError(
            f"Unknown sensor key '{uid}'. "
            "Please check your experiment schedule."
        )

    def get_actuator_definition(self, uid):
        for actdef in self.actuator_definitions:
            if actdef.uid == uid:
                return actdef.actuator_ids

        raise KeyError(
            f"Unknown actuator key '{uid}'. "
            "Please check your experiment schedule."
        )

    @staticmethod
    def load(stream):
        content = load_experiment_file(stream)

        env_definitions: List[EnvironmentDefinition] = []
        for uid, definition in content["definitions"]["environments"].items():
            environment = definition["environment"]
            reward = definition.setdefault("reward", {})
            state_trans = definition.setdefault("state_transformer", {})

            env_definitions.append(
                EnvironmentDefinition(
                    uid=uid,
                    environment_uid=environment.get("uid", "default_env"),
                    environment_name=environment["name"],
                    environment_params=environment.get("params", {}),
                    reward_name=reward.get("name", ""),
                    reward_params=reward.get("params", {}),
                    state_transformer_name=state_trans.get("name", ""),
                    state_transformer_params=state_trans.get("params", {}),
                )
            )

        return Experiment(
            uid=content.get("uid", content.get("id", None)),
            seed=content.get("seed", None),
            version=content.get("version", None),
            repetitions=content.get("repetitions", 1),
            max_runs=content.get("max_runs", -1),
            output_path=content.get(
                "output",
                os.path.abspath(os.path.join(os.getcwd(), "_outputs")),
            ),
            environment_definitions=env_definitions,
            agent_definitions=[
                AgentDefinition(
                    uid=key,
                    name=value.get("name", f"Unnamed Agent {idx}"),
                    load=value.get(
                        "load",
                        {"agent": value.get("name", f"Unnamed Agent {idx}")},
                    ),
                    replay=value.get("replay", []),
                    brain_name=value["brain"]["name"],
                    brain_params=value["brain"].get("params", {}),
                    muscle_name=value["muscle"]["name"],
                    muscle_params=value["muscle"].get("params", {}),
                    objective_name=value["objective"]["name"],
                    objective_params=value["objective"].get("params", {}),
                )
                for idx, (key, value) in enumerate(
                    content["definitions"]["agents"].items()
                )
            ],
            sensor_definitions=[
                SensorDefinition(uid=key, sensor_ids=value)
                for key, value in content["definitions"]["sensors"].items()
            ],
            actuator_definitions=[
                ActuatorDefinition(uid=key, actuator_ids=value)
                for key, value in content["definitions"]["actuators"].items()
            ],
            simulation_definitions=[
                SimulationDefinition(
                    uid=key,
                    simulation_name=value["name"],
                    conditions=[cond for cond in value["conditions"]],
                )
                for key, value in content["definitions"]["simulation"].items()
            ],
            phase_config_definitions=[
                PhaseConfigDefinition(
                    uid=key,
                    mode=value["mode"],
                    worker=value["worker"],
                    episodes=value["episodes"],
                )
                for key, value in content["definitions"][
                    "phase_config"
                ].items()
            ],
            run_config_definition=RunConfigDefinition(
                condition_name=content["definitions"]["run_config"][
                    "condition"
                ]["name"],
                condition_params=content["definitions"]["run_config"][
                    "condition"
                ]["params"],
            ),
            schedule=content["schedule"],
        )



"""This module contains the :class:`ParameterDefinition` that stores
the parameters and their configuration.

"""


class ParameterDefinition:
    """The definition of parameters for the design to create.

    Parameters can either be independent variables, which can change
    over the course of the experiment, or control variables, which stay
    the same for the whole experiment.

    The parameters are used by the generator to generate the design.

    Attributes
    ----------
    required_runs: int
        The number of runs required with the current parameter
        configuration.
    ind_vars: dict
        A *dict* containing IDs of independent variables and their
        configuration.
    ctrl_vars: dict
        A *dict* containing IDs of control variables and their
        configuration.

    """

    def __init__(self):
        self.required_runs = 1
        self.ind_vars = dict()
        self.ctrl_vars = dict()

    def add_independent_variable(self, name, levels):
        """Add an independent variable to the configuration.

        Depending on the number of levels, the required_runs will be
        increased.

        Parameters
        ----------
        name: str
            A unique but readable ID of the variable to add.
        levels: dict
            A *dict* with the configuration of each level.

        """
        self.required_runs *= len(levels)
        self.ind_vars[name] = levels

    def add_control_variable(self, name, level):
        """Add a control variable to the configuration.

        Parameters
        ----------
        name: str
            A unique but readable ID of the variable to add.
        level: dict
            A *dict* containing the configuration of the level.

        """
        self.ctrl_vars[name] = level



import pandas as pd
from arsenai.description import ParameterDefinition
from arsenai.generator.analyzer import analyze_experiment
from arsenai.generator.designer import create_design
from arsenai.generator.experiment import Experiment
from arsenai.schedule.schedule import Schedule

from . import LOG


class Generator:
    """The palaestrAI experiment generator.

    The generator gets the content of an experiment file as an
    :class:`Experiment` object and creates several run files.

    """

    def __init__(self):
        self.required_runs = 1
        self.experiment: Experiment
        self.parameter_def: ParameterDefinition
        self.design: pd.DataFrame
        self.runs: list

    def generate(self, experiment):
        """Generate several runs from one experiment."""
        LOG.debug("Analyzing experiment: %s", experiment)
        self.experiment = experiment
        self.parameter_def = analyze_experiment(experiment)
        if self.parameter_def.required_runs > 1:
            self.design = create_design(
                self.parameter_def, experiment.max_runs
            )
        else:
            self.design = pd.DataFrame({"0._._": [0.0]})

        self.runs = self._build_runs()

        return self.runs, self.design

    def _build_runs(self):
        """Build the run configuration for all required runs.

        The number of runs generated depends on the design (i.e. the
        number of factors and levels or the variable max_runs) and the
        number of repetitions.

        Returns
        -------
        list:
            A list where each entry is a full run configuration.
        """
        runs = list()

        run_idx = 0
        for _ in range(self.experiment.repetitions):
            for idx in range(len(self.design)):
                run = dict()
                run["uid"] = f"{self.experiment.uid}-{run_idx}"
                run["experiment_uid"] = self.experiment.uid
                run["seed"] = self.experiment.get_seed()
                run["version"] = self.experiment.get_version()
                run["run_config"] = self.experiment.get_run_config()
                run["schedule"] = self._build_schedule(idx)
                runs.append(run)
                run_idx += 1
        return runs

    def _build_schedule(self, design_idx):
        """Build the schedule for a specific experiment run.

        This function creates the run configuration for a specific
        configuration of the experiment. All DoE-related information
        in the experiment-file-schedule will be replaced with the
        necessary information for the run file.

        Parameters
        ----------
        idx: int
            The index of the current design configuration.

        Returns
        -------
        schedule: list
            The schedule with all information required by palaestrAI
            to perform an experiment run.

        """
        schedule = Schedule(self.experiment.get_schedule())

        schedule.populate()

        schedule.build(self, design_idx)
        # schedule = self._popuplate_schedule(schedule)

        # for idx, phase in enumerate(schedule):
        #     phase_name = list(phase.keys())[0]
        #     self._build_phase(schedule, idx, phase_name, design_idx)

        # for idx, phase in enumerate(schedule):
        #     for agent_cfg in list(phase.values())[0]["agents"]:
        #         try:
        #             del agent_cfg["uid"]
        #         except KeyError:
        #             pass

        return schedule.as_list()

    # def _popuplate_schedule(self, schedule):
    #     """Generate the full schedule.

    #     The experiment file allows to skip entries in the schedule
    #     that are the same like in the previous phase. This function
    #     copies all information from the previous phase and updates
    #     with the current phase. This is saved as populated schedule.

    #     Parameters
    #     ----------
    #     schedule: list
    #         A list with all phase configs. The configs contain only
    #         minimal information.

    #     Returns
    #     -------
    #     list
    #         A list with all phase configs. The configs contain all
    #         information for that phase.

    #     """
    #     populated = list()

    #     prev_name = ""
    #     prev_config = dict()
    #     for idx, phase in enumerate(schedule):
    #         phase_name = list(phase.keys())[0]
    #         phase_config = phase[phase_name]
    #         if idx > 0:
    #             prev_config = deepcopy(populated[idx - 1][prev_name])
    #         else:
    #             prev_config = dict()

    #         phase_config = update_dict(prev_config, phase_config)
    #         populated.append({phase_name: phase_config})
    #         prev_name = phase_name
    #     return populated

    # def _build_phase(self, schedule, idx, phase_name, design_idx):
    #     """Build the config for a phase.

    #     Takes the input from the schedule defined in the experiment
    #     file and replaces all keys with actual configs. For each key
    #     that is subject of DoE, the corresponding config is identified
    #     with help of the experiment design.

    #     The schedule will be updated inplace.

    #     Parameters
    #     ----------
    #     schedule: list
    #         The populated schedule from the experiment file.
    #     idx: int
    #         The index of the current phase.
    #     phase_name: str
    #         The name of the current phase.
    #     design_idx: int
    #         The index of the current experiment run in the design.

    #     """
    #     phase_def = schedule[idx][phase_name]
    #     if idx > 0:
    #         prev_def = list(schedule[idx - 1].values())[0]
    #     else:
    #         prev_def = dict()

    #     design = self.design.iloc[design_idx]

    #     for factor in [
    #         "environments",
    #         "agents",
    #         "simulation",
    #         "phase_config",
    #     ]:

    #         config = None

    #         for name, level in design.items():
    #             if factor not in name:
    #                 continue

    #             _phase_idx, _, _ = name.split(".")
    #             if int(_phase_idx) != idx:
    #                 continue

    #             user_key = phase_def[factor][int(level)]
    #             config = self.experiment.get_definition(
    #                 factor, user_key, phase_name
    #             )
    #             break

    #         if config is None:
    #             copy_last = False
    #             if any([factor in val for val in design.index]) and prev_def:
    #                 copy_last = True
    #             if len(phase_def[factor]) == 1:
    #                 copy_last = False

    #             if copy_last:
    #                 config = deepcopy(prev_def[factor])
    #             else:
    #                 user_key = phase_def[factor][0]
    #                 config = self.experiment.get_definition(
    #                     factor, user_key, phase_name
    #                 )
    #         phase_def[factor] = config

    #     for factor in ["sensors", "actuators"]:
    #         for agent_key, defs in phase_def[factor].items():
    #             level = 0
    #             for name, dlevel in design.items():
    #                 _phase_idx, _factor, _agent_key = name.split(".")
    #                 if (
    #                     factor != _factor
    #                     or agent_key != _agent_key
    #                     or int(_phase_idx) != idx
    #                 ):
    #                     continue
    #                 level = int(dlevel * (len(phase_def[factor]) - 1))
    #                 break

    #             config = self.experiment.get_definition(
    #                 factor, defs[level], phase_name
    #             )
    #             for agent_cfg in phase_def["agents"]:
    #                 if agent_cfg["uid"] == agent_key:
    #                     agent_cfg[factor] = list()
    #                     for env_cfg in phase_def["environments"]:
    #                         env_uid = [val for val in env_cfg.values()][0][
    #                             "uid"
    #                         ]
    #                         if env_uid not in config:
    #                             continue

    #                         agent_cfg.setdefault(factor, list())
    #                         for uid in config[env_uid]:
    #                             sen_key = f"{env_uid}.{uid}"
    #                             if sen_key not in agent_cfg[factor]:
    #                                 agent_cfg[factor].append(sen_key)
    #     try:
    #         del phase_def["sensors"]
    #     except KeyError:
    #         pass
    #     try:
    #         del phase_def["actuators"]
    #     except KeyError:
    #         pass



#!usr/bin/env python3
# arsenai.py -- Manager CLI for ARL DoE generation for palaestrAI.
# Copyright (C) 2020  OFFIS, All rights reserved.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library.

import sys
import importlib.metadata

import click
import palaestrai
import setproctitle
from palaestrai.cli.manager import init_logger
from palaestrai.core import RuntimeConfig

from arsenai.api import fnc_generate


@click.group()
@click.option(
    "-c",
    "--config",
    type=click.Path(),
    help="Supply custom runtime configuration file. "
    "(Default search path: %s)"
    % (palaestrai.core.runtime_config._RuntimeConfig.CONFIG_FILE_PATHS),
)
@click.option(
    "--verbose",
    "-v",
    count=True,
    help="Increases the program verbosity, can be given numerous times: "
    "-v prints also INFO messages, and -vv emits DEBUG output."
    "output",
)
@click.version_option(importlib.metadata.version("palaestrai-arsenai"))
def main(config=None, verbose=0):
    setproctitle.setproctitle(" ".join(sys.argv))

    if config:
        try:
            with open(config, "r") as fp:
                RuntimeConfig().load(fp)
        except OSError as err:
            click.echo(
                f"ERROR: Could not load config from {config}: {err}.",
                file=sys.stderr,
            )
            exit(1)
    else:
        try:
            RuntimeConfig().load()
        except FileNotFoundError as err:
            click.echo(
                f"Please create a runtime config: {err}.\n"
                "Will continue with built-in defaults.",
                file=sys.stderr,
            )
    init_logger(verbose)


@main.command()
@click.argument(
    "experiment_file",
    type=click.Path(
        exists=True, dir_okay=False, readable=True, allow_dash=True
    ),
)
def generate(experiment_file):
    click.echo(f"Reading experiment file: {experiment_file}.")

    fnc_generate.generate(experiment_file)

    click.echo("ArsenAI finished!")



class Phase:
    """This class stores information of a specific phase in a
    schedule.

    """

    def __init__(self, phase_idx: int, name: str, blueprint: dict):
        self.phase_idx: int = phase_idx
        self.name: str = name
        self.blueprint: dict = blueprint
        self.environments: list = None
        self.agents: list = None
        self.simulation: dict = None
        self.phase_config: dict = None

        self.sensors: dict = None
        self.actuators: dict = None

        self._phase: dict = dict()
        self.requests: list = list()

    def populate(self):
        self.environments = self._read_blueprint("environments")
        self.agents = self._read_blueprint("agents")
        self.simulation = self._read_blueprint("simulation")
        self.phase_config = self._read_blueprint("phase_config")
        self.sensors = self._read_blueprint("sensors")
        self.actuators = self._read_blueprint("actuators")

    def _read_blueprint(self, key):
        content = self.blueprint.get(key, ["$req($prev)"])
        if isinstance(content, list):
            for level_idx, level in enumerate(content):
                if isinstance(level, list):
                    # environments, agents
                    for entry_idx, entry in enumerate(level):
                        request = {}
                        if "$req" in entry:
                            request = {
                                "level": level_idx,
                                "what": key,
                                "entry": entry_idx,
                                "replace": entry,
                            }
                            args = (
                                entry.split("(")[-1].split(")")[0].split(",")
                            )

                            for arg in args:
                                if "phase" in arg:
                                    request["from_phase"] = int(
                                        arg.split("=")[-1]
                                    )
                                elif "index" in arg:
                                    request["index"] = int(arg.split("=")[-1])
                                else:
                                    raise ValueError(
                                        f"Unknown argument: {arg}"
                                    )

                        if request:
                            self.requests.append(request)

                elif isinstance(level, dict):
                    # sensors, actuators
                    pass

                elif isinstance(level, str):
                    # everything else
                    request = {}
                    if "$req" in level:
                        request = {
                            "level": level_idx,
                            "what": key,
                            "replace": level,
                        }
                        args = level.split("(")[-1].split(")")[0].split(",")
                        for arg in args:
                            if arg == "$prev":
                                request["from_phase"] = self.phase_idx - 1
                            elif "phase" in arg:
                                request["from_phase"] = int(arg.split("=")[0])
                            else:
                                raise ValueError(f"Unknown argument: {arg}")
                    if request:
                        self.requests.append(request)
            # content = list()

        # if isinstance(content, str):
        #     # everything else, have no index
        #     if "$req" in content:
        #         args = content.split("(")[-1].split(")")[0]
        #         args = args.split(",")
        #         request = {"index": 0, "what": key}
        #         for arg in args:
        #             if arg == "$prev":
        #                 request["from_phase"] = self.phase_idx - 1
        #             elif "phase" in arg:
        #                 _, target = arg.split("=")
        #                 request["from_phase"] = int(target)
        #             else:
        #                 raise ValueError(f"Unknown argument: {arg}")
        #         self.requests.append(request)
        #     content = list()

        return content

    def build(self, generator, design_idx):
        design = generator.design.iloc[design_idx]

        for factor in [
            "environments",
            "agents",
            "simulation",
            "phase_config",
        ]:
            config = None
            for name, level in design.items():
                if factor not in name:
                    continue
                _phase_idx, _, _ = name.split(".")
                if int(_phase_idx) != self.phase_idx:
                    continue
                user_key = getattr(self, factor)[int(level)]
                config = generator.experiment.get_definition(
                    factor, user_key, self.name
                )
                setattr(self, factor, user_key)
                break

            if config is None:
                user_keys = getattr(self, factor)
                if isinstance(user_keys, list):
                    user_keys = user_keys[0]
                # user_keys = getattr(self, factor)[0]
                if isinstance(user_keys, str):
                    user_keys = [user_keys]

                config = generator.experiment.get_definition(
                    factor, user_keys, self.name
                )
            self._phase[factor] = config

        for factor in ["sensors", "actuators"]:
            for agent_key, defs in getattr(self, factor).items():
                # agent_key = agent_def["uid"]
                level = 0
                for (
                    name,
                    dlevel,
                ) in design.items():
                    _phase_idx, _factor, _agent_key = name.split(".")
                    if (
                        factor != _factor
                        or agent_key != _agent_key
                        or int(_phase_idx) != self.phase_idx
                    ):
                        continue
                    level = int(dlevel * (len(getattr(self, factor)) - 1))
                    break

                config = generator.experiment.get_definition(
                    factor, defs[level], self.name
                )
                for agent_cfg in self._phase["agents"]:
                    if agent_cfg["uid"] == agent_key:
                        agent_cfg[factor] = list()
                        for env_cfg in self._phase["environments"]:
                            env_uid = [val for val in env_cfg.values()][0][
                                "uid"
                            ]
                            if env_uid not in config:
                                continue
                            for uid in config[env_uid]:
                                if not str(uid).startswith(env_uid):
                                    key = f"{env_uid}.{uid}"
                                else:
                                    key = uid
                                if key not in agent_cfg[factor]:
                                    agent_cfg[factor].append(key)
                # if agent
        for agent_cfg in self._phase["agents"]:
            del agent_cfg["uid"]

    def as_dict(self):
        return {self.name: self._phase}



from __future__ import annotations
from typing import List, Dict, Union
from dataclasses import dataclass


@dataclass
class AgentDefinition:
    uid: str
    name: str
    load: dict
    replay: List[Dict[str, Union[str, int]]]
    brain_name: str
    brain_params: dict
    muscle_name: str
    muscle_params: dict
    objective_name: str
    objective_params: dict



"""This module contains the :func:`analyze_experiment` that is used
to determine the number of required runs and to split parameters into
independent and control variables.

"""

from ..description import ParameterDefinition
from . import LOG


def analyze_experiment(experiment):
    """Analyze an experiment.

    Calculates and returns the number of required runs if every
    possible parameter configuration is tested. Also creates a
    parameter definition that stores independent and control variables
    and their configuration.

    Parameters
    ----------
    experiment: Experiment
        The :class:`Experiment` object containing the content of the
        experiment file.

    Returns
    -------
    ParameterDefinition
        The parameter definition created from the analysis of the
        experiment object.

    """
    params = ParameterDefinition()

    LOG.info(
        "Experiment '%s' has %d run phase(s).",
        experiment.uid,
        len(experiment.schedule),
    )
    for idx, phase in enumerate(experiment.schedule):
        if len(phase) > 1:
            LOG.warning(
                "Phase %d has more than one (=%d) phase definitions. "
                "Every definition after the first one will be ignored.",
                idx,
                len(phase),
            )
        phase_def = phase[list(phase.keys())[0]]

        _analyze_phase(idx, phase_def, params)
        _analyze_sensors_and_actuators(idx, phase_def, params)

    LOG.info(
        "Experiment '%s' allows %d different configuration(s).",
        experiment.uid,
        params.required_runs,
    )
    return params


def _analyze_phase(idx, phase_def, params):
    """Analyze entries of a phase.

    Splits the configuration into independent variables and control
    variables.

    Variable keys have the format: idx.factor._, where idx is the index
    of the current phase, factor is one of `environments`, `agents`,
    `simulation`, or `phase_config`. The underscore is not used here
    but it is added to have the same schema for all variables.

    Parameters
    ----------
    idx: int
        Index of the current phase.
    phase_def: dict
        The definition of the current phase from the experiment file.
    params: ParameterDefinition
        The parameter definition where the variables will be added.

    """
    for factor in ["environments", "agents", "simulation", "phase_config"]:
        if factor in phase_def:
            levels = len(phase_def[factor])

            key = f"{idx}.{factor}._"
            param_cfg = dict()
            for lidx, level in enumerate(phase_def[factor]):
                param_cfg[lidx] = level
            if levels > 1:
                params.add_independent_variable(key, param_cfg)
            else:
                params.add_control_variable(key, param_cfg)

            LOG.debug(
                "Phase %d: Factor '%s' has %d level(s).",
                idx,
                factor,
                levels,
            )


def _analyze_sensors_and_actuators(idx, phase_def, params):
    """Analyze sensor and actuator entries of a phase.

    Splits the configuration into independent variables and control
    variables.

    Variable keys have the format: idx.factor.agent_key, where idx is
    the index of the current phase, factor is one of `sensors` or
    `actuators`. Since sensors are defined per-agent, the agent_key
    specifies to which agent the configuration belongs to.

    Parameters
    ----------
    idx: int
        Index of the current phase.
    phase_def: dict
        The definition of the current phase from the experiment file.
    params: ParameterDefinition
        The parameter definition where the variables will be added.

    """
    for factor in ["sensors", "actuators"]:
        if factor not in phase_def:
            continue

        for agent_key, defs in phase_def[factor].items():
            levels = len(defs)

            key = f"{idx}.{factor}.{agent_key}"
            param_cfg = dict()
            for lidx, level in enumerate(defs):
                param_cfg[lidx] = level

            if levels > 1:
                params.add_independent_variable(key, param_cfg)
            else:
                params.add_control_variable(key, param_cfg)

            LOG.debug(
                "Phase %d: Factor '%s' has %d level(s) for agent %s.",
                idx,
                factor,
                levels,
                agent_key,
            )



"""This module contains util functions for reading experiment files and
writing run files.
"""

import collections.abc
import logging
import os

import ruamel
from ruamel.yaml.constructor import ConstructorError

LOG = logging.getLogger(__name__)


def load_experiment_file(stream):
    """Read the content of an experiment file from disk.

    The file to be loaded is expected to be a yaml file and needs to
    have a specific set of keys. See the example in
    `tests/fixtures/example_experiment.yml`.

    Parameters
    ----------
    stream: Union[str, TextIOWrapper]
        A filepath or a stream where to load the experiment from.

    Returns
    -------
    dict
        The content of the experiment file as *dict*.

    """
    LOG.debug("Loading experiment from %s.", stream)

    if isinstance(stream, str):
        try:
            stream = open(stream, "r")
        except OSError as err:
            LOG.error("Could not open experiment file: %s.", err)
            raise err
    try:
        content = ruamel.yaml.YAML(typ="safe", pure=True).load(stream)
    except ConstructorError as err:
        LOG.error("Coud not load experiment file: %s.", err)
        raise err
    finally:
        stream.close()
    LOG.debug("Loaded experiment: %s.", content)
    return content


def create_run_files(runs, experiment_id, output_path):
    """Write a list of run files to disk.

    Each run is a *dict* with a specific set of keys and values. The
    resulting file will be a run file that can be executed without
    further modification by palestrAI core.

    If you have specified an output path in your experiment file, the
    run files will be created at that location. Otherwise, a folder
    *_outputs* will be created in the current working directory.

    Parameters
    ----------
    runs: List[dict]
        A *list* of *dict*s where each dict represents an experiment
        run.
    experiment_id: str
        The `uid` of the experiment. Will be added to the filename
        of the runs.
    output_path: str
        The output path where the run files will be stored. Can be
        modified in the experiment file (key: output).

    """
    os.makedirs(output_path, exist_ok=True)
    yaml = ruamel.yaml.YAML(typ="safe", pure=True)
    for idx, run in enumerate(runs):
        filename = f"{experiment_id}_run-{idx}.yml"
        path = os.path.join(output_path, filename)

        with open(path, "w") as yaml_file:
            yaml.dump(run, yaml_file)


def update_dict(src, upd):
    """Recursive update of dictionaries.

    See stackoverflow:

        https://stackoverflow.com/questions/3232943/
        update-value-of-a-nested-dictionary-of-varying-depth

    """
    for key, val in upd.items():
        if isinstance(val, collections.abc.Mapping):
            src[key] = update_dict(src.get(key, dict()), val)
        else:
            src[key] = val
    return src



"""This module contains the :func:`create_design` that creates an
experiment design for a specific parameter definition.

"""

import doepy.build
import numpy as np
import pandas as pd

from . import LOG


def create_design(parameter_definition, max_runs, n_iter=1000):
    """Create a sampling design.

    If the number of required runs is lower than the maximal number of
    allowed runs, a full-factorial design is created and returned.

    Otherwise, a subset of the full-factorial design will be optimized
    according to the maximin metric.

    Parameters
    ----------
    parameter_definition: ParameterDefinition
        The definition of parameters of the experiment, containing
        independent and control variables.
    max_runs: int
        The maximal number of runs to create. Can be defined in the
        experiment file.
    n_iter: int, optional
        The number of iterations of the optimizer.

    Returns
    -------
    pd.DataFrame
        A pandas DataFrame containing the experiment design for the
        independent variables.

    """
    design = doepy.build.full_fact(parameter_definition.ind_vars)

    if max_runs < 0 or parameter_definition.required_runs <= max_runs:
        return design
    else:
        sample = design.sample(n=max_runs)
        sample_values = sample.values
        sample_optimized = _optimize(sample_values, design.values, n_iter)
        design = pd.DataFrame(
            data=sample_optimized, index=sample.index, columns=sample.columns
        )

    return design


def _optimize(design, original, n_iter):
    """Optimize a design with an evolutionary algorithm (EA).

    The (1+1)-EA uses the maximin metric to optimize the design.

    Parameters
    ----------
    design: np.ndarray
        The initial sampling design.
    original: np.ndarray
        The full-factorial design to sample from.
    n_iter: int
        The number of iterations of the EA.

    Returns
    -------
    np.ndarray
        The final design with the best score according to maximin.

    """

    candidate = design
    current1 = candidate.copy()

    for idx in range(n_iter):
        current2 = candidate.copy()

        # Create offspring
        _mutate(current1, original)
        _mutate(current2, original)

        if _isbetter(current1, candidate):
            candidate = current1
            LOG.debug("Candidate 1 has a better design in iteration %d.", idx)
        if _isbetter(current2, candidate):
            candidate = current2
            LOG.debug("Candidate 2 has a better design in iteration %d.", idx)

    return candidate


def _mutate(design, original):
    """Mutate a given design from the original (full) design.

    Randomly choose a point in the design matrix and mutate the design
    point within the level's range. First, by extracting the ranges of
    the factors from the design, randomly choose a point in the design:
    `design[idx_row][idx_col]`, and, finally, randomize it within the
    range.

    Parameters
    ----------
    design: np.ndarray
        The array of the design to be mutated.
    original: np.ndarray
        The array of the full factorial design.

    """
    high_ranges = np.zeros(design.shape[1])
    low_ranges = np.zeros(design.shape[1])

    for col in range(design.shape[1]):
        high_ranges[col] = np.max(original[:, col])
        low_ranges[col] = np.min(original[:, col])

    # TODO: seed the rng
    idx_row = np.random.randint(design.shape[0])
    idx_col = np.random.randint(design.shape[1])

    design[idx_row][idx_col] = np.random.randint(
        low_ranges[idx_col], high_ranges[idx_col] + 1
    )


def _isbetter(design1, design2, mode="maximin", p=8):
    """Return whether the first design is better than the second.

    The designs are rated based on the maximin metric (see equation 8.9
    in `Statistische Versuchsplanung: Design of Experiments (DoE) by
    Siebertz et al.).

    Notes
    -----
        The p value is chosen experimentally, considering the 0 - 1
        range of the current factor levels, it should be examined
        according to the factors and their ranges to achieve a better
        distribution of the points

    Parameters
    ----------
    design1: np.ndarray
        The first design to rate.
    design2: np.ndarray
        The second design to rate and to compare with.
    mode: str, optional
        The metric to use. Currently, only `maximin` is possible.
    p: int, optional
        The p value used to calculate the maximin.

    Returns
    -------
    bool
        True if design1 is better, False otherwise.

    """
    if mode == "maximin":
        maximinp_design1 = sum(_pdist(design1) ** -p) ** (1 / p)
        maximinp_design2 = sum(_pdist(design2) ** -p) ** (1 / p)
    else:
        raise ValueError(f"Mode {mode} is not supported.")

    return maximinp_design1 < maximinp_design2


def _pdist(design):
    """Calculate the point-wise distance of the design.

    Parameters
    ----------
    design: np.ndarray
        The design for which the distance is calculated.

    Returns
    -------
    np.ndarray
        Point-wise distances of the design's points.

    """
    num, dim = design.shape
    distances = list()

    for idx in range(num - 1):
        for jdx in range(idx + 1, num):
            distances.append(
                sum((design[jdx, :] - design[idx, :]) ** 2) ** 0.5
            )

    return np.array(distances)
